##############################################################################
#
# Copyright (c) 2004 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Security related configuration fields.
"""
__docformat__ = 'restructuredtext'

from zope.configuration.fields import GlobalObject
from zope.configuration.fields import MessageID
from zope.interface import Interface
from zope.schema import Id
from zope.schema.interfaces import IFromUnicode

from zope.security._compat import implementer_if_needed
from zope.security.interfaces import PUBLIC_PERMISSION_NAME as zope_Public
from zope.security.management import setSecurityPolicy
from zope.security.permission import checkPermission


@implementer_if_needed(IFromUnicode)
class Permission(Id):
    r"""This field describes a permission.
    """

    def fromUnicode(self, value):
        u = super().fromUnicode(value)

        map = getattr(self.context, 'permission_mapping', {})
        return map.get(u, u)

    def _validate(self, value):
        super()._validate(value)

        if value != zope_Public:
            self.context.action(
                discriminator=None,
                callable=checkPermission,
                args=(None, value),

                # Delay execution till end. This is an
                # optimization. We don't want to intersperse utility
                # lookup, done when checking permissions, with utility
                # definitions. Utility lookup is expensive after
                # utility definition, as extensive caches have to be
                # rebuilt.
                order=9999999,
            )


class ISecurityPolicyDirective(Interface):
    """Defines the security policy that will be used for Zope."""

    component = GlobalObject(
        title="Component",
        description="Pointer to the object that will handle the security.",
        required=True)


def securityPolicy(_context, component):
    _context.action(
        discriminator='defaultPolicy',
        callable=setSecurityPolicy,
        args=(component,)
    )


class IPermissionDirective(Interface):
    """Define a new security object."""

    id = Id(
        title="ID",
        description="ID as which this object will be known and used.",
        required=True)

    title = MessageID(
        title="Title",
        description="Provides a title for the object.",
        required=True)

    description = MessageID(
        title="Description",
        description="Provides a description for the object.",
        required=False)


def permission(_context, id, title, description=''):
    from zope.component.zcml import utility

    from zope.security.interfaces import IPermission
    from zope.security.permission import Permission
    permission = Permission(id, title, description)
    utility(_context, IPermission, permission, name=id)


class IRedefinePermission(Interface):
    """Define a permission to replace another permission."""

    from_ = Permission(
        title="Original permission",
        description="Original permission ID to redefine.",
        required=True)

    to = Permission(
        title="Substituted permission",
        description="Substituted permission ID.",
        required=True)


def redefinePermission(_context, from_, to):
    _context = _context.context

    # check if context has any permission mappings yet
    if not hasattr(_context, 'permission_mapping'):
        _context.permission_mapping = {}

    _context.permission_mapping[from_] = to



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Test handler for 'protectClass' directive
"""
import unittest

from zope.security.interfaces import PUBLIC_PERMISSION_NAME as zope_Public


class Test_protectName(unittest.TestCase):

    def setUp(self):
        from zope.security.checker import _clear
        _clear()

    def tearDown(self):
        from zope.security.checker import _clear
        _clear()

    def _callFUT(self, class_, name, permission):
        from zope.security.protectclass import protectName
        return protectName(class_, name, permission)

    def test_wo_existing_checker_w_zope_Public(self):
        from zope.security.checker import CheckerPublic
        from zope.security.checker import _checkers
        self._callFUT(Foo, 'bar', zope_Public)
        self.assertIs(_checkers[Foo].get_permissions['bar'], CheckerPublic)

    def test_w_existing_checker(self):
        from zope.security.checker import Checker
        from zope.security.checker import _checkers
        checker = _checkers[Foo] = Checker({})
        permission = object()
        self._callFUT(Foo, 'bar', permission)
        self.assertIs(_checkers[Foo], checker)
        self.assertIs(checker.get_permissions['bar'], permission)


class Test_protectSetAttribute(unittest.TestCase):

    def setUp(self):
        from zope.security.checker import _clear
        _clear()

    def tearDown(self):
        from zope.security.checker import _clear
        _clear()

    def _callFUT(self, class_, name, permission):
        from zope.security.protectclass import protectSetAttribute
        return protectSetAttribute(class_, name, permission)

    def test_wo_existing_checker_w_zope_Public(self):
        from zope.security.checker import CheckerPublic
        from zope.security.checker import _checkers
        self._callFUT(Foo, 'bar', zope_Public)
        self.assertIs(_checkers[Foo].set_permissions['bar'], CheckerPublic)

    def test_w_existing_checker(self):
        from zope.security.checker import Checker
        from zope.security.checker import _checkers
        checker = _checkers[Foo] = Checker({})
        permission = object()
        self._callFUT(Foo, 'bar', permission)
        self.assertIs(_checkers[Foo], checker)
        self.assertIs(checker.set_permissions['bar'], permission)


class Test_protectLikeUnto(unittest.TestCase):

    def setUp(self):
        from zope.security.checker import _clear
        _clear()

    def tearDown(self):
        from zope.security.checker import _clear
        _clear()

    def _callFUT(self, class_, like_unto):
        from zope.security.protectclass import protectLikeUnto
        return protectLikeUnto(class_, like_unto)

    def test_wo_existing_like_unto_checker(self):
        from zope.security.checker import _checkers
        self.assertNotIn(Foo, _checkers)
        self._callFUT(Bar, Foo)
        self.assertNotIn(Foo, _checkers)
        self.assertNotIn(Bar, _checkers)

    def test_w_existing_like_unto_checker_wo_existing_checker(self):
        from zope.security.checker import Checker
        from zope.security.checker import CheckerPublic
        from zope.security.checker import _checkers
        from zope.security.checker import defineChecker
        permission = object()
        foo_checker = Checker({'bar': CheckerPublic}, {'bar': permission})
        defineChecker(Foo, foo_checker)
        self._callFUT(Bar, Foo)
        bar_checker = _checkers[Bar]
        self.assertEqual(bar_checker.get_permissions,
                         foo_checker.get_permissions)
        self.assertEqual(bar_checker.set_permissions,
                         foo_checker.set_permissions)

    def test_w_existing_like_unto_checker_w_existing_checker(self):
        from zope.security.checker import Checker
        from zope.security.checker import CheckerPublic
        from zope.security.checker import _checkers
        from zope.security.checker import defineChecker
        permission1, permission2 = object(), object()
        foo_checker = Checker({'bar': CheckerPublic}, {'bar': permission2})
        defineChecker(Foo, foo_checker)
        bar_checker = Checker({'bar': permission1, 'baz': CheckerPublic}, {})
        defineChecker(Bar, bar_checker)
        self._callFUT(Bar, Foo)
        bar_checker = _checkers[Bar]
        self.assertEqual(bar_checker.get_permissions,
                         {'bar': CheckerPublic, 'baz': CheckerPublic})
        self.assertEqual(bar_checker.set_permissions,
                         foo_checker.set_permissions)


class Foo:
    bar = 'Bar'


class Bar(Foo):
    baz = 'Baz'


def test_suite():
    return unittest.TestSuite((
        unittest.defaultTestLoader.loadTestsFromTestCase(Test_protectName),
        unittest.defaultTestLoader.loadTestsFromTestCase(
            Test_protectSetAttribute),
        unittest.defaultTestLoader.loadTestsFromTestCase(Test_protectLikeUnto),
    ))



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""A small sandbox application.
"""
import random
import time

from zope.interface import Interface
from zope.interface import implementer


class IAgent(Interface):
    """A player/agent in the world.

    The agent represents an autonomous unit, that lives in various
    homes/sandboxes and accesses services present at the sandboxes. Agents are
    imbued with a sense of wanderlust and attempt to find new homes after a
    few turns of the time generator (think turn based games).
    """
    def action():
        """Perform agent's action."""

    def setHome(home):
        """Move to a different home."""

    def getHome():
        """Return the place where the agent currently lives."""

    def getAuthenticationToken():
        """Return the authority by which the agent perform actions."""


class IService(Interface):
    """Marker to designate some form of functionality.

    Services are available from sandboxes, examples include time service,
    agent discovery, and sandbox discovery.
    """


class ISandbox(Interface):
    """A container for agents to live in and services to be available."""

    def getService(service_id):
        """Get the service having the provided id in this sandbox."""

    def getAgents():
        """Return a list of agents living in this sandbox."""

    def addAgent(agent):
        """Add a new agent to the sandbox."""

    def transportAgent(agent, destination):
        """Move the specified agent to the destination sandbox."""


class SandboxError(Exception):
    """A sandbox error is thrown, if any action could not be performed."""
    pass


class Identity:
    """Mixin for pretty printing and identity method"""

    def __init__(self, id, *args, **kw):
        self.id = id

    def getId(self):
        return self.id

    def __str__(self):
        return f"<{str(self.__class__.__name__)}> {str(self.id)}"

    __repr__ = __str__


@implementer(IAgent)
class Agent(Identity):

    def __init__(self, id, home, auth_token, action):
        """Initialize agent."""
        self.id = id
        self.auth_token = auth_token
        self.home = home
        self._action = action

    def action(self):
        """See IAgent."""
        self._action(self, self.getHome())

    def setHome(self, home):
        """See IAgent."""
        self.home = home

    def getHome(self):
        """See IAgent."""
        return self.home

    def getAuthenticationToken(self):
        """See IAgent."""
        return self.auth_token


@implementer(ISandbox)
class Sandbox(Identity):
    """
    see ISandbox doc
    """

    def __init__(self, id, service_factories):
        self.id = id
        self._services = {}
        self._agents = {}

        for sf in service_factories:
            self.addService(sf())

    def getAgentIds(self):
        return self._agents.keys()

    def getAgents(self):
        return self._agents.values()

    def getServiceIds(self):
        return self._services.keys()

    def getService(self, sid):
        return self._services.get(sid)

    def getHome(self):
        return self

    def addAgent(self, agent):
        if agent.getId() not in self._agents \
                and IAgent.providedBy(agent):
            self._agents[agent.getId()] = agent
            agent.setHome(self)
        else:
            raise SandboxError("couldn't add agent %s" % agent)

    def addService(self, service):

        if not service.getId() in self._services \
                and IService.providedBy(service):
            self._services[service.getId()] = service
            service.setHome(self)
        else:
            raise SandboxError("couldn't add service %s" % service)

    def transportAgent(self, agent, destination):
        if agent.getId() in self._agents \
                and destination is not self \
                and ISandbox.providedBy(destination):
            destination.addAgent(agent)
            del self._agents[agent.getId()]
        else:
            raise SandboxError("couldn't transport agent {} to {}".format(
                agent, destination)
            )


@implementer(IService)
class Service:
    def getId(self):
        return self.__class__.__name__

    def setHome(self, home):
        self._home = home

    def getHome(self):
        return getattr(self, '_home')


class HomeDiscoveryService(Service):
    """
    returns the ids of available agent homes
    """

    def getAvailableHomes(self):
        return _homes.keys()


class AgentDiscoveryService(Service):
    """
    returns the agents available at a given home
    """

    def getLocalAgents(self, home):
        return home.getAgents()


class TimeService(Service):
    """
    returns the local time
    """

    def getTime(self):
        return time.time()


default_service_factories = (
    HomeDiscoveryService,
    AgentDiscoveryService,
    TimeService
)


def action_find_homes(agent, home):
    home_service = home.getService('HomeDiscoveryService')
    return home_service.getAvailableHomes()


def action_find_neighbors(agent, home):
    agent_service = home.getService('AgentDiscoveryService')
    return agent_service.getLocalAgents(home)


def action_find_time(agent, home):
    time_service = home.getService('TimeService')
    return time_service.getTime()


class TimeGenerator:
    """Represents the passage of time in the agent simulation.

    each turn represents some discrete unit of time, during
    which all agents attempt to perform their action. Additionally,
    all agents are checked to see if they have a desire to move,
    and if so are transported to a new random home.
    """

    def setupAgent(self, agent):
        pass

    def teardownAgent(self, agent):
        pass

    def turn(self):
        for h in _homes.values():
            agents = h.getAgents()
            for a in agents:
                self.setupAgent(a)
                try:
                    a.action()
                except Exception as e:
                    print('-- Exception --')
                    print('"%s" in "%s" not allow to "%s"'
                          % (a, h, a._action.__name__))
                    print(e)
                    print()
                self.teardownAgent(a)

            agents = filter(WanderLust, agents)

            for a in agents:
                self.setupAgent(a)
                try:
                    home = a.getHome()
                    new_home = GreenerPastures(a)
                    home.transportAgent(a, new_home)
                except Exception as e:
                    print('-- Exception --')
                    print('moving "%s" from "%s" to "%s"' %
                          (a, h, repr(new_home)))
                    print(e)
                    print()
                self.teardownAgent(a)


def WanderLust(agent):
    """ is agent ready to move """
    if int(random.random() * 100) <= 30:
        return 1


def GreenerPastures(agent):
    """ where do they want to go today """
    possible_homes = _homes.keys()
    possible_homes.remove(agent.getHome().getId())
    return _homes.get(random.choice(possible_homes))


# boot strap initial setup.

# global list of homes
_homes = {}

all_homes = (
    Sandbox('jail', default_service_factories),
    Sandbox('origin', default_service_factories),
    Sandbox('valhalla', default_service_factories)
)

origin = all_homes[1]

for h in all_homes:
    _homes[h.getId()] = h


agents = [
    Agent('odin', None, 'norse legend', action_find_time),
    Agent('loki', None, 'norse legend', action_find_neighbors),
    Agent('thor', None, 'norse legend', action_find_homes),
    Agent('thucydides', None, 'greek men', action_find_time),
    Agent('archimedes', None, 'greek men', action_find_neighbors),
    Agent('prometheus', None, 'greek men', action_find_homes),
]

for a in agents:
    origin.addAgent(a)


def main():
    world = TimeGenerator()

    for x in range(5):
        print('world turning')
        world.turn()

    for h in _homes.values():
        print(h.getId(), h.getAgentIds())


if __name__ == '__main__':
    main()



##############################################################################
#
# Copyright (c) 2004 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Test permissions
"""
import unittest

from zope.component.testing import PlacelessSetup

from zope.security.interfaces import PUBLIC_PERMISSION_NAME as zope_Public


class PermissionTests(unittest.TestCase):

    def _getTargetClass(self):
        from zope.security.permission import Permission
        return Permission

    def _makeOne(self, id, *args):
        klass = self._getTargetClass()
        return klass(id, *args)

    def test_class_conforms_to_IPermission(self):
        from zope.interface.verify import verifyClass

        from zope.security.interfaces import IPermission
        verifyClass(IPermission, self._getTargetClass())

    def test_instance_conforms_to_IPermission(self):
        from zope.interface.verify import verifyObject
        from zope.schema import getValidationErrors

        from zope.security.interfaces import IPermission
        verifyObject(IPermission, self._makeOne('testing'))
        self.assertEqual([],
                         getValidationErrors(IPermission,
                                             self._makeOne('testing')))

    def test_ctor_only_id(self):
        permission = self._makeOne('testing')
        self.assertEqual(permission.id, 'testing')
        self.assertEqual(permission.title, '')
        self.assertEqual(permission.description, '')

    def test_ctor_w_title_and_description(self):
        permission = self._makeOne('testing', 'TITLE', 'DESCRIPTION')
        self.assertEqual(permission.id, 'testing')
        self.assertEqual(permission.title, 'TITLE')
        self.assertEqual(permission.description, 'DESCRIPTION')


class Test_checkPermission(PlacelessSetup, unittest.TestCase):

    def _callFUT(self, context, permission_id):
        from zope.security.permission import checkPermission
        return checkPermission(context, permission_id)

    def test_w_CheckerPublic(self):
        from zope.security.checker import CheckerPublic
        self._callFUT(None, CheckerPublic)  # no raise

    def test_miss(self):
        self.assertRaises(ValueError, self._callFUT, None, 'nonesuch')

    def test_hit(self):
        from zope.component import provideUtility

        from zope.security.interfaces import IPermission
        permission = object()
        provideUtility(permission, IPermission, 'testing')
        self._callFUT(None, 'testing')  # no raise


class Test_allPermissions(PlacelessSetup, unittest.TestCase):

    def _callFUT(self):
        from zope.security.permission import allPermissions
        return allPermissions()

    def test_empty(self):
        self.assertEqual(list(self._callFUT()), [])

    def test_w_registration(self):
        self.assertEqual(list(self._callFUT()), [])
        from zope.component import provideUtility

        from zope.security.interfaces import IPermission
        permission = object()
        provideUtility(permission, IPermission, 'testing')
        self.assertEqual(list(self._callFUT()), ['testing'])

    def test_skips_zope_Public(self):
        self.assertEqual(list(self._callFUT()), [])
        from zope.component import provideUtility

        from zope.security.checker import CheckerPublic
        from zope.security.interfaces import IPermission
        permission = object()
        provideUtility(permission, IPermission, 'testing')
        provideUtility(CheckerPublic, IPermission, zope_Public)
        self.assertEqual(list(self._callFUT()), ['testing'])


class Test_PermissionsVocabulary(PlacelessSetup, unittest.TestCase):

    def _callFUT(self):
        from zope.security.permission import PermissionsVocabulary
        return PermissionsVocabulary()

    def test_empty(self):
        from zope.schema.vocabulary import SimpleVocabulary
        vocabulary = self._callFUT()
        self.assertIsInstance(vocabulary, SimpleVocabulary)
        self.assertEqual(list(vocabulary), [])

    def test_w_registration(self):
        self.assertEqual(list(self._callFUT()), [])
        from zope.component import provideUtility

        from zope.security.interfaces import IPermission
        permission = object()
        provideUtility(permission, IPermission, 'testing')
        vocabulary = self._callFUT()
        self.assertEqual([x.token for x in vocabulary], ['testing'])

    def test_includes_zope_Public(self):
        self.assertEqual(list(self._callFUT()), [])
        from zope.component import provideUtility

        from zope.security.checker import CheckerPublic
        from zope.security.interfaces import IPermission
        permission = object()
        provideUtility(permission, IPermission, 'testing')
        provideUtility(CheckerPublic, IPermission, zope_Public)
        vocabulary = self._callFUT()
        self.assertEqual(sorted([x.token for x in vocabulary]),
                         ['testing', zope_Public])

    def test_zcml_valid(self):
        from zope.configuration import xmlconfig
        from zope.interface.verify import verifyObject
        from zope.schema import getValidationErrors

        import zope.security
        from zope.security.interfaces import IPermission

        xmlconfig.file('configure.zcml', zope.security)
        vocabulary = self._callFUT()
        vocabulary = sorted(vocabulary, key=lambda term: term.token)
        self.assertEqual(6, len(vocabulary))

        for term in vocabulary:
            p = term.value
            __traceback_info__ = term.token, p
            verifyObject(IPermission, p)
            self.assertEqual([], getValidationErrors(IPermission, p))


class Test_PermissionIdsVocabulary(PlacelessSetup, unittest.TestCase):

    def _callFUT(self):
        from zope.security.permission import PermissionIdsVocabulary
        return PermissionIdsVocabulary()

    def test_empty(self):
        from zope.schema.vocabulary import SimpleVocabulary
        vocabulary = self._callFUT()
        self.assertIsInstance(vocabulary, SimpleVocabulary)
        self.assertEqual(list(vocabulary), [])

    def test_w_registration(self):
        self.assertEqual(list(self._callFUT()), [])
        from zope.component import provideUtility

        from zope.security.interfaces import IPermission
        permission = object()
        provideUtility(permission, IPermission, 'testing')
        vocabulary = self._callFUT()
        self.assertEqual([x.value for x in vocabulary], ['testing'])
        self.assertEqual([x.token for x in vocabulary], ['testing'])

    def test_includes_zope_Public(self):
        self.assertEqual(list(self._callFUT()), [])
        from zope.component import provideUtility

        from zope.security.checker import CheckerPublic
        from zope.security.interfaces import IPermission
        permission = object()
        provideUtility(permission, IPermission, 'testing')
        provideUtility(CheckerPublic, IPermission, zope_Public)
        vocabulary = self._callFUT()
        self.assertEqual([x.value for x in vocabulary],
                         [CheckerPublic, 'testing'])
        self.assertEqual([x.token for x in vocabulary],
                         [zope_Public, 'testing'])


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Security Checkers.

This module contains the primary implementations of
:class:`zope.security.interfaces.IChecker` (:class:`Checker`,
:class:`MultiChecker`, :func:`NamesChecker`) and
:class:`zope.security.interfaces.IProxyFactory` (:func:`ProxyFactory`).

It also defines helpers for permission checking (:func:`canAccess`,
:func:`canWrite`) and getting checkers
(:func:`getCheckerForInstancesOf`, :func:`selectChecker`).

This module is accelerated with a C implementation on CPython by
default. If the environment variable ``PURE_PYTHON`` (converted to
``int``) has a nonzero value before this module is imported, the C
extensions will be bypassed and the reference Python implementations
will be used. This can be helpful for debugging and tracing.

Debugging Permissions Problems
==============================

You can set the environment variable ``ZOPE_WATCH_CHECKERS`` before
this module is imported to get additional security checker debugging
output on the standard error.

Setting ``ZOPE_WATCH_CHECKERS`` to 1 will display messages about unauthorized
or forbidden attribute access.  Setting it to a larger number will also display
messages about granted attribute access.

Note that the ``ZOPE_WATCH_CHECKERS`` mechanism may eventually be
replaced with a more general security auditing mechanism.

.. seealso:: :class:`CheckerLoggingMixin`
.. seealso:: :class:`WatchingChecker`
.. seealso:: :class:`WatchingCombinedChecker`

API
===

.. py:data:: CheckerPublic

  The special constant that indicates that no permission
  checking needs to be done.

.. autofunction:: selectChecker
"""
import abc
import datetime
import decimal
import io
import os
import sys
import types
import weakref

import zope.interface.declarations
import zope.interface.interface
import zope.interface.interfaces
from zope.i18nmessageid import Message
from zope.interface import Interface
from zope.interface import directlyProvides
from zope.interface import implementer
from zope.interface.interfaces import IDeclaration
from zope.interface.interfaces import IInterface

from zope.security._compat import PURE_PYTHON
from zope.security._compat import implementer_if_needed
from zope.security._definitions import thread_local
from zope.security.interfaces import ForbiddenAttribute
from zope.security.interfaces import IChecker
from zope.security.interfaces import INameBasedChecker
from zope.security.interfaces import ISecurityProxyFactory
from zope.security.interfaces import Unauthorized
from zope.security.proxy import Proxy
from zope.security.proxy import getChecker


try:
    from zope.exceptions import DuplicationError
except ImportError:  # pragma: no cover
    class DuplicationError(Exception):
        """A duplicate registration was attempted"""

WATCH_CHECKERS = 0

if os.environ.get('ZOPE_WATCH_CHECKERS'):  # pragma: no cover
    try:
        WATCH_CHECKERS = int(os.environ.get('ZOPE_WATCH_CHECKERS'))
    except ValueError:
        WATCH_CHECKERS = 1


def ProxyFactory(object, checker=None):
    """Factory function that creates a proxy for an object

    The proxy checker is looked up if not provided.
    """
    if isinstance(object, Proxy):
        if checker is None or checker is getChecker(object):
            return object
        else:
            # We have a proxy, but someone asked us to change its checker.
            # Let's raise an exception.
            #
            # Other reasonable actions would be to either keep the existing
            # proxy, or to create a new one with the given checker.
            # The latter might be a security hole though, if untrusted code
            # can call ProxyFactory.
            raise TypeError("Tried to use ProxyFactory to change a Proxy's"
                            " checker.")
    if checker is None:
        checker = getattr(object, '__Security_checker__', None)

        if checker is None:
            checker = selectChecker(object)
            if checker is None:
                return object

    return Proxy(object, checker)


directlyProvides(ProxyFactory, ISecurityProxyFactory)

# This import represents part of the API for the proxy module
from . import proxy  # noqa: E402 module level import not at top


proxy.ProxyFactory = ProxyFactory


def canWrite(obj, name):
    """Check whether the interaction may write an attribute named name on obj.

    Convenience method.  Rather than using checkPermission in high level code,
    use canWrite and canAccess to avoid binding code to permissions.
    """
    obj = ProxyFactory(obj)
    checker = getChecker(obj)
    try:
        checker.check_setattr(obj, name)
    except Unauthorized:
        return False
    except ForbiddenAttribute:
        # we are going to be a bit DWIM-y here: see
        # http://www.zope.org/Collectors/Zope3-dev/506

        # generally, if the check is ForbiddenAttribute we want it to be
        # raised: it probably indicates a programming or configuration error.
        # However, we special case a write ForbiddenAttribute when one can
        # actually read the attribute: this represents a reasonable
        # configuration of a readonly attribute, and returning False (meaning
        # "no, you can't write it") is arguably more useful than raising the
        # exception.
        try:
            checker.check_getattr(obj, name)
            # we'll let *this* ForbiddenAttribute fall through, if any.  It
            # means that both read and write are forbidden.
        except Unauthorized:
            pass
        return False
    # all other exceptions, other than Unauthorized and ForbiddenAttribute,
    # should be passed through uncaught, as they indicate programmer error
    return True


def canAccess(obj, name):
    """Check whether the interaction may access an attribute named name on obj.

    Convenience method.  Rather than using checkPermission in high level code,
    use canWrite and canAccess to avoid binding code to permissions.
    """
    # access attributes and methods, including, in the current checker
    # implementation, special names like __getitem__
    obj = ProxyFactory(obj)
    checker = getChecker(obj)
    try:
        checker.check_getattr(obj, name)
    except Unauthorized:
        return False
    # if it is Forbidden (or anything else), let it be raised: it probably
    # indicates a programming or configuration error
    return True


@implementer(INameBasedChecker)
class CheckerPy:
    """
    The Python reference implementation of
    :class:`zope.security.interfaces.INameBasedChecker`.

    Ordinarily there will be no reason to ever explicitly use this class;
    instead use the class assigned to :class:`Checker`.
    """

    def __init__(self, get_permissions, set_permissions=None):
        """Create a checker

        A dictionary must be provided for computing permissions for
        names. The dictionary get will be called with attribute names
        and must return a permission ID, None, or the special marker,
        :const:`CheckerPublic`. If None is returned, then access to the name is
        forbidden. If :const:`CheckerPublic` is returned, then access will be
        granted without checking a permission.

        An optional setattr dictionary may be provided for checking
        set attribute access.

        """
        if not isinstance(get_permissions, dict):
            raise TypeError('get_permissions must be a dict')
        self.get_permissions = get_permissions
        if set_permissions is not None:
            if not isinstance(set_permissions, dict):
                raise TypeError('set_permissions must be a dict')
        else:
            set_permissions = {}
        self.set_permissions = set_permissions

    def permission_id(self, name):
        'See INameBasedChecker'
        return self.get_permissions.get(name)

    def setattr_permission_id(self, name):
        'See INameBasedChecker'
        if self.set_permissions:
            return self.set_permissions.get(name)

    def check_setattr(self, object, name):
        'See IChecker'
        if self.set_permissions:
            permission = self.set_permissions.get(name)
        else:
            permission = None

        if permission is not None:
            if permission is CheckerPublic:
                return  # Public
            if thread_local.interaction.checkPermission(permission, object):
                return  # allowed
            else:
                __traceback_supplement__ = (TracebackSupplement, object)
                raise Unauthorized(object, name, permission)

        __traceback_supplement__ = (TracebackSupplement, object)
        raise ForbiddenAttribute(name, object)

    def check(self, object, name):
        'See IChecker'
        permission = self.get_permissions.get(name)
        if permission is not None:
            if permission is CheckerPublic:
                return  # Public
            if thread_local.interaction.checkPermission(permission, object):
                return
            else:
                __traceback_supplement__ = (TracebackSupplement, object)
                raise Unauthorized(object, name, permission)
        elif name in _available_by_default:
            return

        if name != '__iter__' or hasattr(object, name):
            __traceback_supplement__ = (TracebackSupplement, object)
            raise ForbiddenAttribute(name, object)

    check_getattr = check  # 'See IChecker'

    def proxy(self, value):
        'See IChecker'
        if isinstance(value, Proxy):
            return value
        checker = getattr(value, '__Security_checker__', None)
        if checker is None:
            checker = selectChecker(value)
            if checker is None:
                return value

        return Proxy(value, checker)


Checker = CheckerPy  # in case no C optimizations


# Helper class for __traceback_supplement__
class TracebackSupplement:

    def __init__(self, obj):
        self.obj = obj

    def getInfo(self):
        result = []
        try:
            cls = self.obj.__class__
            if hasattr(cls, "__module__"):
                s = f"{cls.__module__}.{cls.__name__}"
            else:  # pragma: no cover XXX
                s = str(cls.__name__)
            result.append("   - class: " + s)
        except:  # pragma: no cover # noqa: E722 do not use bare 'except'
            pass
        try:
            cls = type(self.obj)
            if hasattr(cls, "__module__"):
                s = f"{cls.__module__}.{cls.__name__}"
            else:  # pragma: no cover XXX
                s = str(cls.__name__)
            result.append("   - type: " + s)
        except:  # pragma: no cover # noqa: E722 do not use bare 'except'
            pass
        return "\n".join(result)


class Global:
    """A global object that behaves like a string.

    We want this to behave as a global, meaning it's pickled
    by name, rather than value. We need to arrange that it has a suitable
    __reduce__.
    """

    def __init__(self, name, module=None):
        if module is None:  # pragma: no cover XXX
            module = sys._getframe(1).f_locals['__name__']

        self.__name__ = name
        self.__module__ = module

    def __reduce__(self):
        return self.__name__

    def __repr__(self):
        return "{}({},{})".format(self.__class__.__name__,
                                  self.__name__, self.__module__)


CheckerPublic = Global('CheckerPublic')
CP_HACK_XXX = CheckerPublic

# Now we wrap it in a security proxy so that it retains its
# identity when it needs to be security proxied.
# XXX: This means that we can't directly document it with
# sphinx because issubclass() will fail.
d = {}
CheckerPublic = Proxy(CheckerPublic, Checker(d))  # XXX uses CheckerPy
d['__reduce__'] = CheckerPublic
d['__module__'] = CheckerPublic
del d

# TODO: It's a bit scary above that we can pickle a proxy if access is
# granted to __reduce__. We might want to bother to prevent this in
# general and only allow it in this specific case.


def NamesChecker(names=(), permission_id=CheckerPublic, **__kw__):
    """Return a checker that grants access to a set of names.

    A sequence of names is given as the first argument. If a second
    argument, permission_id, is given, it is the permission required
    to access the names.  Additional names and permission IDs can be
    supplied as keyword arguments.
    """

    data = {}
    data.update(__kw__)
    for name in names:
        if data.get(name, permission_id) is not permission_id:
            raise DuplicationError(name)
        data[name] = permission_id

    return Checker(data)


def InterfaceChecker(interface, permission_id=CheckerPublic, **__kw__):
    """
    Create a :func:`NamesChecker` for all the names defined in the *interface*
    (a subclass of :class:`zope.interface.Interface`).
    """
    return NamesChecker(interface.names(all=True), permission_id, **__kw__)


def MultiChecker(specs):
    """
    Create a checker from a sequence of specifications

    A specification is:

    - A two-tuple with:

      o a sequence of names or an interface

      o a permission ID

      All the names in the sequence of names or the interface are
      protected by the permission.

    - A dictionary (having an items method), with items that are
      name/permission-id pairs.
    """
    data = {}

    for spec in specs:
        if isinstance(spec, tuple):
            names, permission_id = spec
            if IInterface.providedBy(names):
                names = names.names(all=True)
            for name in names:
                if data.get(name, permission_id) is not permission_id:
                    raise DuplicationError(name)
                data[name] = permission_id
        else:
            for name, permission_id in spec.items():
                if data.get(name, permission_id) is not permission_id:
                    raise DuplicationError(name)
                data[name] = permission_id

    return Checker(data)


def selectCheckerPy(object):
    """Get a checker for the given object

    The appropriate checker is returned or None is returned. If the
    return value is None, then object should not be wrapped in a proxy.
    """

    # We need to be careful here. We might have a proxy, in which case
    # we can't use the type.  OTOH, we might not be able to use the
    # __class__ either, since not everything has one.

    # TODO: we really need formal proxy introspection

    # if type(object) is Proxy:
    #    # Is this already a security proxy?
    #    return None

    checker = _getChecker(type(object), _defaultChecker)

    # checker = _getChecker(getattr(object, '__class__', type(object)),
    #                      _defaultChecker)

    if checker is NoProxy:
        return None

    while not isinstance(checker, Checker):
        checker = checker(object)

        if checker is NoProxy or checker is None:
            return None

    return checker


selectChecker = selectCheckerPy  # in case no C optimizations


def getCheckerForInstancesOf(class_):
    return _checkers.get(class_)


DEFINABLE_TYPES = (type, types.ModuleType)


def defineChecker(type_, checker):
    """Define a checker for a given type of object

    The checker can be a :class:`Checker`, or a function that, when called with
    an object, returns a :class:`Checker`.
    """
    if not isinstance(type_, DEFINABLE_TYPES):
        raise TypeError(
            'type_ must be a type, class or module, not a %s' % type_)
    if type_ in _checkers:
        raise DuplicationError(type_)
    _checkers[type_] = checker


def undefineChecker(type_):
    del _checkers[type_]


NoProxy = object()

# _checkers is a mapping.
#
#  - Keys are types
#
#  - Values are
#
#    o None => rock
#    o a Checker
#    o a function returning None or a Checker
#
_checkers = {}

_defaultChecker = Checker({})
_available_by_default = []

# Get optimized versions
_c_available = not PURE_PYTHON
if _c_available:  # pragma: no cover
    try:
        import zope.security._zope_security_checker
    except (ImportError, AttributeError):
        _c_available = False

if _c_available:  # pragma: no cover
    from zope.security._zope_security_checker import Checker
    from zope.security._zope_security_checker import NoProxy
    from zope.security._zope_security_checker import _available_by_default
    from zope.security._zope_security_checker import _checkers
    from zope.security._zope_security_checker import _defaultChecker
    from zope.security._zope_security_checker import selectChecker
    zope.interface.classImplements(Checker, INameBasedChecker)


_getChecker = _checkers.get


@implementer_if_needed(IChecker)
class CombinedChecker(Checker):
    """A checker that combines two other checkers in a logical-or fashion.

    The following table describes the result of a combined checker in detail.

    +--------------------+--------------------+-------------------------------------+
    | checker1           | checker2           | CombinedChecker(checker1, checker2) |
    +====================+====================+=====================================+
    | ok                 | anything           | ok (checker 2 never called)         |
    +--------------------+--------------------+-------------------------------------+
    | Unauthorized       | ok                 | ok                                  |
    +--------------------+--------------------+-------------------------------------+
    | Unauthorized       | Unauthorized       | Unauthorized                        |
    +--------------------+--------------------+-------------------------------------+
    | Unauthorized       | ForbiddenAttribute | Unauthorized                        |
    +--------------------+--------------------+-------------------------------------+
    | ForbiddenAttribute | ok                 | ok                                  |
    +--------------------+--------------------+-------------------------------------+
    | ForbiddenAttribute | Unauthorized       | Unauthorized                        |
    +--------------------+--------------------+-------------------------------------+
    | ForbiddenAttribute | ForbiddenAttribute | ForbiddenAttribute                  |
    +--------------------+--------------------+-------------------------------------+
    """  # noqa: E501 line too long

    def __init__(self, checker1, checker2):
        """Create a combined checker."""
        Checker.__init__(self,
                         checker1.get_permissions,
                         checker1.set_permissions)

        self._checker2 = checker2

    def check(self, object, name):
        'See IChecker'
        try:
            Checker.check(self, object, name)
        except ForbiddenAttribute:
            self._checker2.check(object, name)
        except Unauthorized as unauthorized_exception:
            try:
                self._checker2.check(object, name)
            except ForbiddenAttribute:
                raise unauthorized_exception

    check_getattr = __setitem__ = check

    def check_setattr(self, object, name):
        'See IChecker'
        try:
            Checker.check_setattr(self, object, name)
        except ForbiddenAttribute:
            self._checker2.check_setattr(object, name)
        except Unauthorized as unauthorized_exception:
            try:
                self._checker2.check_setattr(object, name)
            except ForbiddenAttribute:
                raise unauthorized_exception


class CheckerLoggingMixin:
    """
    Debugging mixin for checkers.

    Prints verbose debugging information about every performed check to
    :data:`sys.stderr`.

    """

    #: If set to 1 (the default), only displays ``Unauthorized`` and
    #: ``Forbidden`` messages. If verbosity is set to a larger number,
    #: displays all messages. Normally this is controlled via the environment
    #: variable ``ZOPE_WATCH_CHECKERS``.
    verbosity = 1
    _file = sys.stderr

    def _log(self, msg, verbosity=1):
        if self.verbosity >= verbosity:
            self._file.write('%s\n' % msg)

    def check(self, object, name):
        try:
            super().check(object, name)
            if self.verbosity > 1:
                if name in _available_by_default:
                    self._log('[CHK] + Always available: %s on %r'
                              % (name, object), 2)
                else:
                    self._log(
                        f'[CHK] + Granted: {name} on {object!r}', 2)
        except Unauthorized:
            self._log(
                f'[CHK] - Unauthorized: {name} on {object!r}')
            raise
        except ForbiddenAttribute:
            self._log(
                f'[CHK] - Forbidden: {name} on {object!r}')
            raise

    def check_getattr(self, object, name):
        try:
            super().check(object, name)
            if self.verbosity > 1:
                if name in _available_by_default:
                    self._log(
                        '[CHK] + Always available getattr: %s on %r'
                        % (name, object), 2)
                else:
                    self._log(
                        '[CHK] + Granted getattr: %s on %r'
                        % (name, object), 2)
        except Unauthorized:
            self._log(
                f'[CHK] - Unauthorized getattr: {name} on {object!r}'
            )
            raise
        except ForbiddenAttribute:
            self._log(
                f'[CHK] - Forbidden getattr: {name} on {object!r}')
            raise

    __setitem__ = check_getattr

    def check_setattr(self, object, name):
        try:
            super().check_setattr(object, name)
            if self.verbosity > 1:
                self._log(
                    '[CHK] + Granted setattr: {} on {!r}'.format(
                        name, object), 2)
        except Unauthorized:
            self._log(
                '[CHK] - Unauthorized setattr: {} on {!r}'.format(
                    name, object))
            raise
        except ForbiddenAttribute:
            self._log(
                f'[CHK] - Forbidden setattr: {name} on {object!r}')
            raise


# We have to be careful with the order of inheritance
# here. See https://github.com/zopefoundation/zope.security/issues/8
class WatchingChecker(CheckerLoggingMixin, Checker):
    """
    A checker that will perform verbose logging. This will be set
    as the default when ``ZOPE_WATCH_CHECKERS`` is set when this
    module is imported.
    """
    verbosity = WATCH_CHECKERS


class WatchingCombinedChecker(CombinedChecker, WatchingChecker):
    """
    A checker that will perform verbose logging. This will be set
    as the default when ``ZOPE_WATCH_CHECKERS`` is set when this
    module is imported.
    """
    verbosity = WATCH_CHECKERS


if WATCH_CHECKERS:  # pragma: no cover
    # When we make these the default, we also need to be sure
    # to update the _defaultChecker's type (if it's not the C
    # extension) so that selectCheckerPy can properly recognize
    # it as a Checker.
    # See https://github.com/zopefoundation/zope.security/issues/8
    Checker = WatchingChecker
    CombinedChecker = WatchingCombinedChecker

    if not _c_available:
        _defaultChecker.__class__ = Checker


def _instanceChecker(inst):
    return _checkers.get(inst.__class__, _defaultChecker)


def moduleChecker(module):
    """
    Return the :class:`zope.security.interfaces.IChecker` defined for the
    *module*, if any.

    .. seealso:: :func:`zope.security.metaconfigure.protectModule`
       To define module protections.
    """
    return _checkers.get(module)


_available_by_default[:] = [
    '__lt__', '__le__', '__eq__',
    '__gt__', '__ge__', '__ne__',
    '__hash__', '__bool__',
    '__class__', '__providedBy__', '__implements__',
    '__repr__', '__conform__',
    '__name__', '__parent__',
]

_callableChecker = NamesChecker(['__str__', '__name__', '__call__'])
_typeChecker = NamesChecker([
    '__str__', '__name__', '__module__', '__bases__', '__mro__',
    '__implemented__',
])
_namedChecker = NamesChecker(['__name__'])
_iteratorChecker = NamesChecker([
    'next', '__next__', '__iter__', '__len__',
    '__length_hint__',
])

_setChecker = NamesChecker([
    '__iter__', '__len__', '__str__', '__contains__',
    'copy', 'difference', 'intersection', 'issubset',
    'issuperset', 'symmetric_difference', 'union',
    '__and__', '__or__', '__sub__', '__xor__',
    '__rand__', '__ror__', '__rsub__', '__rxor__',
    '__eq__', '__ne__', '__lt__', '__gt__',
    '__le__', '__ge__',
])


class _BasicTypes(dict):
    """Basic Types Dictionary

    Make sure that checkers are really updated, when a new type is added.
    """

    def __setitem__(self, name, value):
        dict.__setitem__(self, name, value)
        _checkers[name] = value

    def __delitem__(self, name):
        dict.__delitem__(self, name)
        del _checkers[name]

    def clear(self):
        # Make sure you cannot clear the values
        raise NotImplementedError

    def update(self, d):
        dict.update(self, d)
        _checkers.update(d)


_basic_types = {
    object: NoProxy,
    int: NoProxy,
    float: NoProxy,
    complex: NoProxy,
    type(None): NoProxy,
    str: NoProxy,
    bytes: NoProxy,
    Message: NoProxy,  # Messages are immutable, so it's okay
    bool: NoProxy,
    datetime.timedelta: NoProxy,
    datetime.datetime: NoProxy,
    datetime.date: NoProxy,
    datetime.time: NoProxy,
    datetime.tzinfo: NoProxy,
    type({}.values()): NoProxy,
    type({}.keys()): NoProxy,
    type({}.items()): NoProxy,
}

try:
    import pytz
except ImportError:  # pragma: no cover
    pass
else:
    _basic_types[type(pytz.UTC)] = NoProxy

BasicTypes = _BasicTypes(_basic_types)
del _basic_types

# Available for tests. Located here so it can be kept in sync with BasicTypes.
BasicTypes_examples = {
    object: object(),
    int: 65536,
    float: -1.4142,
    complex: -1.4142j,
    type(None): None,
    bytes: b'abc',
    bool: True,
    datetime.timedelta: datetime.timedelta(3),
    datetime.datetime: datetime.datetime(2003, 1, 1),
    datetime.date: datetime.date(2003, 1, 1),
    datetime.time: datetime.time(23, 58),
    Message: Message('message', domain='hello')
}


class _Sequence:
    def __len__(self):
        raise NotImplementedError()

    def __getitem__(self, i):
        raise NotImplementedError()


_Declaration_checker = InterfaceChecker(
    IDeclaration,
    _implied=CheckerPublic,
    subscribe=CheckerPublic,
    unsubscribe=CheckerPublic,
    __call__=CheckerPublic,
)


def f():  # pragma: no cover
    yield f


_default_checkers = {
    dict: NamesChecker(['__getitem__', '__len__', '__iter__',
                        'get', 'has_key', 'copy', '__str__', 'keys',
                        'values', 'items', 'iterkeys', 'iteritems',
                        'itervalues', '__contains__']),
    list: NamesChecker(['__getitem__', '__len__', '__iter__',
                        '__contains__', 'index', 'count', '__str__',
                        '__add__', '__radd__', ]),
    set: _setChecker,
    frozenset: _setChecker,
    # XXX: actually decimal.Decimal has more methods, which are unlisted here
    #      so expect ForbiddenAttribute on such
    decimal.Decimal: NamesChecker(['__bool__', '__cmp__', '__eq__',
                                   '__ne__', '__hash__',
                                   '__str__',
                                   '__neg__', '__pos__', '__abs__',
                                   '__add__', '__radd__',
                                   '__sub__', '__rsub__',
                                   '__mul__', '__rmul__',
                                   '__truediv__', '__rtruediv__',
                                   '__divmod__', '__rdivmod__',
                                   '__mod__', '__rmod__',
                                   '__floordiv__', '__rfloordiv__',
                                   '__float__', '__int__',
                                   '__pow__', '__rpow__',
                                   'adjusted', 'as_tuple', 'compare',
                                   'max', 'min', 'normalize',
                                   'quantize', 'remainder_near',
                                   'same_quantum', 'sqrt',
                                   'to_eng_string', 'to_integral']),

    # YAGNI: () a rock
    tuple: NamesChecker(['__getitem__', '__add__', '__radd__',
                         '__contains__', '__len__', '__iter__',
                         '__str__']),
    Proxy: NoProxy,
    type(weakref.ref(_Sequence())): NamesChecker(['__call__']),
    types.FunctionType: _callableChecker,
    types.MethodType: _callableChecker,
    types.BuiltinFunctionType: _callableChecker,
    types.BuiltinMethodType: _callableChecker,
    # At least on Python 3.5-3.12, types.BuiltinFunctionType and
    # types.BuiltinMethodType are both <class 'builtin_function_or_method'>,
    # or PyCFunctionType.  However, some builtin methods are <class
    # 'builtin_method'> instead, or PyCMethodType, which is a subclass of
    # PyCFunctionType but not identical to it.  As of Python 3.12, the io
    # module makes more use of PyCMethodType, so we can use it to identify
    # that type.
    type(io.BytesIO().getbuffer): _callableChecker,
    # method-wrapper
    type(().__repr__): _callableChecker,
    type: _typeChecker,
    types.ModuleType: lambda module: _checkers.get(module, _namedChecker),
    type(iter([])): _iteratorChecker,
    type(iter(())): _iteratorChecker,
    type(iter({})): _iteratorChecker,
    type(iter(set())): _iteratorChecker,
    type(iter(_Sequence())): _iteratorChecker,
    type(f()): _iteratorChecker,
    type(Interface): InterfaceChecker(
        IInterface,
        __str__=CheckerPublic,
        _implied=CheckerPublic,
        subscribe=CheckerPublic,
        # To iterate, Python calls __len__ as a hint.
        # AttributeErrors are passed.
        __len__=CheckerPublic,
    ),
    zope.interface.interface.Method: InterfaceChecker(
        zope.interface.interfaces.IMethod),
    zope.interface.declarations.ProvidesClass: _Declaration_checker,
    zope.interface.declarations.ClassProvides: _Declaration_checker,
    zope.interface.declarations.Implements: _Declaration_checker,
    zope.interface.declarations.Declaration: _Declaration_checker,
    abc.ABCMeta: _typeChecker,
}


def _fixup_dictlike(dict_type):
    empty_dict = dict_type()
    populated_dict = dict_type({1: 2})
    for dictlike in (empty_dict, populated_dict):
        for attr in ('__iter__', 'keys', 'items', 'values'):
            obj = getattr(dictlike, attr)()
            o_type = type(obj)
            if o_type not in _default_checkers:
                _default_checkers[o_type] = _iteratorChecker
            # PyPy3 has special types for iter({}.items()) etc.
            iter_type = type(iter(obj))
            if iter_type not in _default_checkers:
                _default_checkers[iter_type] = _iteratorChecker


def _fixup_odict():
    from collections import OrderedDict

    # The `_fixup_dictlike` is detected as undefined because it is deleted
    # later on but this function is called beforehand:
    _fixup_dictlike(OrderedDict)  # noqa: F821 undefined name '_fixup_dictlike'


_fixup_odict()
del _fixup_odict

try:
    import BTrees  # noqa: F401 'BTrees' imported but unused
except ImportError:  # pragma: no cover
    pass
else:
    # The C implementation of BTree.items() is its own iterator
    # and doesn't need any special entries to enable iteration.
    # But the Python implementation has to call __iter__ to be able
    # to do iteration. Whitelist it so that they behave the same.
    # In addition, Python 3 will attempt to call __len__ on iterators
    # for a length hint, so the C implementations also need to be
    # added to the _iteratorChecker. The same thing automatically
    # applies for .keys() and .values() since they return the same type.
    # We do this here so that all users of zope.security can benefit
    # without knowing implementation details.
    # See https://github.com/zopefoundation/zope.security/issues/20

    def _fixup_btrees():
        import BTrees._base
        _default_checkers[BTrees._base._TreeItems] = _iteratorChecker

        for name in ('IF', 'II', 'IO', 'OI', 'OO'):
            for family_name in ('family32', 'family64'):
                family = getattr(BTrees, family_name)
                btree = getattr(family, name).BTree
                # The `_fixup_dictlike` is detected as undefined because it is
                # deleted later on but this function is called beforehand:
                _fixup_dictlike(btree)  # noqa: F821 undefined name

    _fixup_btrees()
    del _fixup_btrees

del _fixup_dictlike


def _fixup_zope_interface():
    # Make sure the provided and implementedBy objects
    # can be iterated.
    # Note that we DO NOT use the _iteratorChecker, but instead
    # we use NoProxy to be sure that the results (of iteration or not) are not
    # proxied. On Python 2, these objects are builtin and don't go through the
    # checking process at all, much like BTrees, so NoProxy is necessary for
    # compatibility. On Python 3, prior to this, iteration was simply not
    # allowed.
    from zope.interface import alsoProvides
    from zope.interface import providedBy

    class I1(Interface):
        pass

    class I2(Interface):
        pass

    @implementer(I1)
    class Obj:
        pass

    o = Obj()

    # This will be the zope.interface.implementedBy from the class
    # a zope.interface.declarations.Implements
    _default_checkers[type(providedBy(o))] = NoProxy

    alsoProvides(o, I2)
    # This will be the zope.interface.Provides from the instance
    _default_checkers[type(providedBy(o))] = NoProxy


_fixup_zope_interface()
del _fixup_zope_interface


def _fixup_itertools():
    # itertools.groupby is a built-in custom iterator type introduced
    # in Python 2.4. It should have the same checker as other built-in
    # iterators.

    # Also, itertools._grouper also needs to be exposed as an
    # iterator. Its type is not exposed by name, but can be accessed
    # like so: type(list(itertools.groupby([0]))[0][1])

    import itertools

    group = itertools.groupby([0])
    type_group = type(group)
    if type_group not in _default_checkers:
        _default_checkers[type_group] = _iteratorChecker

    type_grouper = type(list(group)[0][1])
    if type_grouper not in _default_checkers:
        _default_checkers[type_grouper] = _iteratorChecker

    # There are also many other custom types in itertools that need the
    # same treatment. See a similar list in
    # test_checker.py:test_itertools_checkers
    def pred(x):
        return x
    iterable = (1, 2, 3)
    pred_iterable = (pred, iterable)
    for func, args in (
            ('count', ()),
            ('cycle', ((),)),
            ('dropwhile', pred_iterable),
            ('islice', (iterable, 2)),
            ('permutations', (iterable,)),
            ('product', (iterable,)),
            ('repeat', (1, 2)),
            ('starmap', pred_iterable),
            ('takewhile', pred_iterable),
            ('tee', (iterable,)),
            ('zip_longest', (iterable,)),
            ('accumulate', (iterable,)),
            ('compress', (iterable, ())),
            ('combinations', (iterable, 1)),
            ('combinations_with_replacement', (iterable, 1)),
    ):
        func = getattr(itertools, func)

        result = func(*args)
        if func == itertools.tee:
            result = result[0]
        tresult = type(result)
        if tresult not in _default_checkers:
            _default_checkers[tresult] = _iteratorChecker


_fixup_itertools()
del _fixup_itertools


def _clear():
    _checkers.clear()
    _checkers.update(_default_checkers)
    _checkers.update(BasicTypes)


_clear()

try:
    from zope.testing.cleanup import addCleanUp
except ImportError:  # pragma: no cover
    pass
else:
    addCleanUp(_clear)



##############################################################################
#
# Copyright (c) 2004-2011 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Testing support code.

This module provides some helper/stub objects for setting up interactions.
"""
import contextlib

import zope.security.management
from zope import component
from zope import interface
from zope.security import interfaces
from zope.security.interfaces import PUBLIC_PERMISSION_NAME
from zope.security.permission import Permission


@interface.implementer(interfaces.IPrincipal)
class Principal:
    """
    A trivial implementation of :class:`zope.security.interfaces.IPrincipal`.
    """

    def __init__(self, id, title=None, description='', groups=None):
        self.id = id
        self.title = title or id
        self.description = description
        if groups is not None:
            self.groups = groups
            interface.directlyProvides(self, interfaces.IGroupAwarePrincipal)


@interface.implementer(interfaces.IParticipation)
class Participation:
    """
    A trivial implementation of
    :class:`zope.security.interfaces.IParticipation`.
    """

    def __init__(self, principal):
        self.principal = principal
        self.interaction = None


def addCheckerPublic():
    """
    Add the CheckerPublic permission as :data:`zope.Public
    <zope.security.interfaces.PUBLIC_PERMISSION_NAME>`.
    """

    perm = Permission(
        PUBLIC_PERMISSION_NAME,
        'Public',
        """Special permission used for resources that are always public

        The public permission is effectively an optimization, sine
        it allows security computation to be bypassed.
        """
    )
    gsm = component.getGlobalSiteManager()
    gsm.registerUtility(perm, interfaces.IPermission, perm.id)

    return perm


def create_interaction(principal_id, **kw):
    """
    Create a new interaction for the given principal ID, make it the
    :func:`current interaction
    <zope.security.management.newInteraction>`, and return the
    :class:`Principal` object.
    """
    principal = Principal(principal_id, **kw)
    participation = Participation(principal)
    zope.security.management.newInteraction(participation)
    return principal


@contextlib.contextmanager
def interaction(principal_id, **kw):
    """
    A context manager for running an interaction for the given
    principal ID.
    """
    if zope.security.management.queryInteraction():
        # There already is an interaction. Great. Leave it alone.
        yield
    else:
        principal = create_interaction(principal_id, **kw)
        try:
            yield principal
        finally:
            zope.security.management.endInteraction()



##############################################################################
#
# Copyright (c) 2004 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Interfaces for security machinery.

These can be categorized into a few different groups of related objects.

* Exceptions

  - :class:`IUnauthorized`
  - :class:`IForbidden`
  - :class:`IForbiddenAttribute`
  - :class:`NoInteraction`

* Utilities

  - :class:`ISecurityManagement`
  - :class:`ISecurityChecking`
  - :class:`ISecurityProxyFactory`
  - :class:`IChecker`
  - :class:`INameBasedChecker`
  - :class:`ISecurityPolicy`

* Principals

  - :class:`IInteraction`
  - :class:`IParticipation`
  - :class:`IInteractionManagement`
  - :class:`IPrincipal`
  - :class:`ISystemPrincipal`
  - :class:`IGroupAwarePrincipal`
  - :class:`IGroupClosureAwarePrincipal`
  - :class:`IGroup`
  - :class:`IMemberGetterGroup`
  - :class:`IMemberAwareGroup`
  - :class:`IPermission`

Anywhere that an API is documented as accepting a permission, it
means the name of the permission, or the special object
:class:`zope.security.checker.CheckerPublic`.
"""

from zope.interface import Attribute
from zope.interface import Interface
from zope.interface import implementer
from zope.interface.common.interfaces import IAttributeError
from zope.interface.common.interfaces import IException
from zope.schema import NativeStringLine
from zope.schema import Text
from zope.schema import TextLine

from zope.security.i18n import ZopeMessageFactory as _


#: The name (id) of the registered :class:`IPermission` utility that signifies
#: that the protected attribute is public.
#:
#: .. versionadded:: 4.2.0
PUBLIC_PERMISSION_NAME = 'zope.Public'


class IUnauthorized(IException):
    """
    The action is not authorized.

    Implemented in :class:`Unauthorized`.
    """


@implementer(IUnauthorized)
class Unauthorized(Exception):
    """
    Some user wasn't allowed to access a resource.

    Default implementation of :class:`IUnauthorized`.
    """


class IForbidden(IException):
    """
    A resource cannot be accessed under any circumstances

    Implemented in :class:`Forbidden`.
    """


@implementer(IForbidden)
class Forbidden(Exception):
    """
    A resource cannot be accessed under any circumstances

    Default implementation if :class:`IForbidden`.
    """


class IForbiddenAttribute(IForbidden, IAttributeError):
    """
    An attribute is unavailable because it is forbidden (private).

    Implemented in :class:`ForbiddenAttribute`.
    """


@implementer(IForbiddenAttribute)
class ForbiddenAttribute(Forbidden, AttributeError):
    """
    An attribute is unavailable because it is forbidden (private).

    Default implementation of :class:`IForbiddenAttribute`.
    """


class ISecurityManagement(Interface):
    """
    Public security management API.

    This is implemented by :mod:`zope.security.management`.
    """

    def getSecurityPolicy():
        """Get the system default security policy."""

    def setSecurityPolicy(aSecurityPolicy):
        """Set the system default security policy.

        This method should only be called by system startup code.  It
        should never, for example, be called during a web request.
        """


class ISecurityChecking(Interface):
    """
    Public security API.
    """

    def checkPermission(permission, object, interaction=None):
        """
        Return whether security policy allows permission on object.

        :param str permission: The permission name.
        :param object: The object being accessed according to the permission.
        :keyword interaction: An :class:`IInteraction`, providing access to
            information such as authenticated principals.  If it is None, the
            current interaction is used.
        """


class ISecurityProxyFactory(Interface):
    """
    A factory for creating security-proxied objects.

    See :class:`zope.security.checker.ProxyFactory` for the
    default implementation.
    """

    def __call__(object, checker=None):
        """
        Create a security proxy

        If a checker (:class:`IChecker`) is given, then use it,
        otherwise, try to figure out a checker.

        If the object is already a security proxy, then it will be
        returned.
        """


class IChecker(Interface):
    """
    Security-proxy plugin objects that implement low-level checks.

    The checker is responsible for creating proxies for
    operation return values, via the ``proxy`` method.

    There are :meth:`check_getattr` and :meth:`check_setattr` methods
    for checking getattr and setattr, and a :meth:`check` method for all
    other operations.

    The check methods will raise errors if access is not allowed.
    They return no value.

    Example (for ``__getitem__``)::

           checker.check(ob, "__getitem__")
           return checker.proxy(ob[key])

    .. seealso:: :mod:`zope.security.checker`
    """

    def check_getattr(ob, name):
        """
        Check whether attribute access is allowed.

        If a checker implements ``__setitem__``, then ``__setitem__``
        will be called rather than ``check`` to ascertain whether an
        operation is allowed. This is a hack that allows significantly
        greater performance due to the fact that low-level operator
        access is much faster than method access.

        :raises: :class:`Unauthorized`
        :raises: :class:`Forbidden`
        :return: Nothing
        """

    def check_setattr(ob, name):
        """
        Check whether attribute assignment is allowed.

        If a checker implements ``__setitem__``, then ``__setitem__``
        will be called rather than ``check`` to ascertain whether an
        operation is allowed. This is a hack that allows significantly
        greater performance due to the fact that low-level operator
        access is much faster than method access.

        :raises: :class:`Unauthorized`
        :raises: :class:`Forbidden`
        :return: Nothing
        """

    def check(ob, operation):
        """
        Check whether *operation* is allowed.

        The operation name is the Python special method name,
        e.g. "__getitem__".

        May raise Unauthorized or Forbidden.  Returns no value.

        If a checker implements ``__setitem__``, then ``__setitem__``
        will be called rather than ``check`` to ascertain whether an
        operation is allowed. This is a hack that allows significantly
        greater performance due to the fact that low-level operator
        access is much faster than method access.

        :raises: :class:`Unauthorized`
        :raises: :class:`Forbidden`
        :return: Nothing
        """

    def proxy(value):
        """
        Return a security proxy for the *value*.

        If a checker implements ``__getitem__``, then ``__getitem__``
        will be called rather than ``proxy`` to proxy the value. This
        is a hack that allows significantly greater performance due to
        the fact that low-level operator access is much faster than
        method access.
        """


class INameBasedChecker(IChecker):
    """
    Security checker that uses permissions to check attribute
    access.
    """

    def permission_id(name):
        """
        Return the permission used to check attribute access on *name*.

        This permission is used by both :meth:`check` and
        :meth:`check_getattr`.
        """

    def setattr_permission_id(name):
        """
        Return the permission used to check attribute assignment on *name*.

        This permission is used by :meth:`check_setattr`.
        """


class ISecurityPolicy(Interface):
    """
    A factory to get :class:`IInteraction` objects.

    .. seealso:: :mod:`zope.security.simplepolicies`
       For default implementations.
    """

    def __call__(participation=None):
        """
        Creates and returns a new :class:`IInteraction` for a given
        request.

        If *participation* is not None, it is added to the new interaction.
        """


class IInteraction(Interface):
    """
    A representation of an interaction between some actors and the
    system.
    """

    participations = Attribute("""An iterable of participations.""")

    def add(participation):
        """Add a participation."""

    def remove(participation):
        """Remove a participation."""

    def checkPermission(permission, object):
        """Return whether security context allows permission on object.

        :param str permission: A permission name
        :param object: The object being accessed according to the permission
        :return: Whether the access is allowed or not.
        :rtype: bool
        """


class IParticipation(Interface):
    """
    A single participant in an interaction.
    """

    interaction = Attribute("The interaction")
    principal = Attribute("The authenticated :class:`IPrincipal`")


class NoInteraction(Exception):
    """No interaction started
    """


class IInteractionManagement(Interface):
    """
    Interaction management API.

    Every thread has at most one active interaction at a time.

    .. seealso:: :mod:`zope.security.management`
       That module provides the default implementation.
    """

    def newInteraction(participation=None):
        """
        Start a new interaction.

        If *participation* is not None, it is added to the new interaction.

        Raises an error if the calling thread already has an interaction.
        """

    def queryInteraction():
        """
        Return the current interaction.

        Return None if there is no interaction.
        """

    def getInteraction():
        """
        Return the current interaction.

        :raise NoInteraction: if there isn't a current interaction.
        """

    def endInteraction():
        """
        End the current interaction.

        Does nothing if there is no interaction.
        """


class IPrincipal(Interface):
    """
    Principals are security artifacts that execute actions in a
    security environment.

    The most common examples of principals include user and group
    objects.

    It is likely that ``IPrincipal`` objects will have associated
    views used to list principals in management interfaces. For
    example, a system in which other meta-data are provided for
    principals might extend ``IPrincipal`` and register a view for the
    extended interface that displays the extended information.
    """

    id = TextLine(
        title=_("Id"),
        description=_("The unique identification of the principal."),
        required=True,
        readonly=True)

    title = TextLine(
        title=_("Title"),
        description=_("The title of the principal. "
                      "This is usually used in the UI."),
        required=False)

    description = Text(
        title=_("Description"),
        description=_("A detailed description of the principal."),
        required=False)


class ISystemPrincipal(IPrincipal):
    """
    A principal that represents the system (application) itself.

    Typically a system principal is granted extra capabilities
    or excluded from certain checks. End users should *not* be able
    to act as the system principal.

    Because speed is often a factor, a single instance of a system principal
    is found at ``zope.security.management.system_user`` and can
    be compared for by identity (e.g., ``if principal is system_user:``).
    """


class IGroupAwarePrincipal(IPrincipal):
    """
    Group aware principal interface.

    Extends ``IPrincipal`` to contain direct group information.
    """

    groups = Attribute(
        'An iterable of :class:`IGroup` objects to which the principal'
        ' directly belongs')


class IGroupClosureAwarePrincipal(IGroupAwarePrincipal):
    """
    A group-aware principal that can recursively flatten the membership
    of groups to return all the groups.
    """

    allGroups = Attribute(
        "An iterable of the full closure of the principal's groups.")


class IGroup(IPrincipal):
    """
    Group of principals
    """


class IMemberGetterGroup(IGroup):
    """
    A group that can get its members.
    """

    def getMembers():
        """Return an iterable of the members of the group"""


class IMemberAwareGroup(IMemberGetterGroup):
    """
    A group that can both set and get its members.
    """

    def setMembers(value):
        """
        Set members of group to the principal IDs in the iterable
        *value*.
        """


class IPermission(Interface):
    """A permission object.

    Note that the ZCML ``<permission>`` directive restricts the ``id`` to
    be an identifier (a dotted name or a URI), but this interface allows
    any native string.
    """

    id = NativeStringLine(
        title=_("Id"),
        description=_("Id as which this permission will be known and used."),
        readonly=True,
        required=True)

    title = TextLine(
        title=_("Title"),
        description=_("Provides a title for the permission."),
        required=True)

    description = Text(
        title=_("Description"),
        description=_("Provides a description for the permission."),
        required=False)



##############################################################################
#
# Copyright (c) 2013 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Python 2 / 3 compatibility
"""
import os
import platform


py_impl = getattr(platform, 'python_implementation', lambda: None)
PYPY = py_impl() == 'PyPy'
PURE_PYTHON = int(os.environ.get('PURE_PYTHON', PYPY))


class implementer_if_needed:
    # Helper to make sure we don't redundantly implement interfaces
    # already inherited. Doing so tends to produce problems with the
    # C3 order. Even though here we could easily statically determine
    # if we need the interface or not, this is used for clarity, to
    # reduce the testing load, and to insulate against changes in
    # super classes.
    def __init__(self, *ifaces):
        self._ifaces = ifaces

    def __call__(self, cls):
        from zope.interface import implementedBy
        from zope.interface import implementer

        ifaces_needed = []
        implemented = implementedBy(cls)
        ifaces_needed = [
            iface
            for iface in self._ifaces
            if not implemented.isOrExtends(iface)
        ]
        return implementer(*ifaces_needed)(cls)
